#include <nds.h>
#include <nds/arm9/input.h>
#include <nds/arm9/console.h>
#include <stdio.h>

typedef void(*call3)(u32,void*,u32);
void read_nvram(u32 src, void *dst, u32 size) {
	((call3)0x2437)(src,dst,size);
}

u8 writeread(u8 data) {
	while (REG_SPICNT & SPI_BUSY);
	REG_SPIDATA = data;
	while (REG_SPICNT & SPI_BUSY);
	return REG_SPIDATA;
}

int flash_verify(u8 *src,u32 dst) {
	int i;
	u8 tmp[256];
	static int result=2;
	
	read_nvram(dst,tmp,256);
	for(i=0;i<256;i++) {
		if(tmp[i]!=src[i])
			break;
	}
	if(i==256)
		return result;
	
	result = 1;
	//write enable
	REG_SPICNT = SPI_ENABLE|SPI_CONTINUOUS|SPI_DEVICE_NVRAM;
	writeread(6);
	REG_SPICNT = 0;
	
	//Wait for Write Enable Latch to be set
	REG_SPICNT = SPI_ENABLE|SPI_CONTINUOUS|SPI_DEVICE_NVRAM;
	writeread(5);
	while((writeread(0)&0x02)==0); //Write Enable Latch
	REG_SPICNT = 0;
		
		

	 
	//page write
	REG_SPICNT = SPI_ENABLE|SPI_CONTINUOUS|SPI_DEVICE_NVRAM;
	writeread(0x0A);
	writeread((dst&0xff0000)>>16);
	writeread((dst&0xff00)>>8);
	writeread(0);
	for (i=0; i<256; i++) {
		writeread(src[i]);
	}

	REG_SPICNT = 0;
	// wait programming to finish
	REG_SPICNT = SPI_ENABLE|SPI_CONTINUOUS|SPI_DEVICE_NVRAM;
	writeread(0x05);
	while(writeread(0)&0x01);	//Write In Progress
	REG_SPICNT = 0;

	//read back & compare
	read_nvram(dst,tmp,256);
	for(i=0;i<256;i++) {
		if(tmp[i]!=src[i])
			return 0;
	}
	return 1;
}

#define SCREENBUFFER 0x2300000
void text(int line, char *s) {
	u16 *p=(u16*)(SCREENBUFFER+line*64);
	while(*s) *p++=*s++;
}

void display_progress(int count, int total, int hide) {
	u16 *progress=(u16*)(SCREENBUFFER+3*64);
	int percent;
	if(total != 0)
		percent = (count * 100) / total;
	else
		percent = 100;
	if(hide == 0)
	{
		if((count > 0) && (percent < 1))
			percent = 1;
		*progress++=(percent/100)+'0';
		percent %= 100;
		*progress++=(percent/10)+'0';
		percent %= 10;
		*progress++=(percent)+'0';
		*progress++='%';
	}
	else
	{
		*progress++=' ';
		*progress++=' ';
		*progress++=' ';
		*progress++=' ';
	}
}

#undef IPC
#define IPC ((u32*)0x27ff000)
int main () {
	u32 key;
	int i, size, result;
	u8 *fw=(u8*)0x2200000;
	int progress=0,progress_total=0;
	
	//get firmware header for arm9
	read_nvram(0, fw, 0x200);

	//wait for arm9
	IPC[0]=*(u32*)&fw[8];
	IPC[1]=(IPC[0] ^ 0xFFFFFFFF);
	while(IPC[1]!=IPC[0])
		swiDelay(0x10000);

	text(0,"Get ready to short SL1.");
	text(1,"Hold L+R+Start to start.");

	do {
		swiDelay(0x10000);
		key=(~REG_KEYINPUT)&0x3ff;
	} while(key!=(KEY_L|KEY_R|KEY_START));

	text(1,"DO NOT shut off until told!");

	//show progress...
	size=(IPC[2]+0xff)&~0xff;
	size=(size+IPC[4]+0xff)&~0xff;

	for(i=0;i<size;i+=256)
		progress_total++;
	display_progress(progress,progress_total,0);
		
	//flash bootstrap (protected)
	text(2,"Writing Failsafe/Bootstrap");
	size=IPC[2];
	for(i=0;i<size;) {
		result = flash_verify(fw+i,i);
		if(result == 1) {
			i+=0x100;
			display_progress(++progress,progress_total,0);
		} else if (result == 2) {
			i+=0x100;
			display_progress(progress,--progress_total,0);
		} else {
			
			display_progress(progress,progress_total,0);
			swiDelay(0x10000);
			if(progress!=0)
			{
				display_progress(progress,progress_total,1);
				swiDelay(0x10000);
			}
		}
	}
	text(2,"Writing Main Firmware     ");

	//flash everything else
	size=IPC[4];
	fw=(u8*)IPC[3];
	for(i=0;i<size;) {
		result = flash_verify(fw+i,i+0x10000);
		if(result == 1) {
			i+=0x100;
			display_progress(++progress,progress_total,0);
		} else if (result == 2) {
			i+=0x100;
			display_progress(progress,--progress_total,0);
		} else {
			display_progress(progress,progress_total,0);
			swiDelay(0x10000);
			if(progress!=0)
			{
				display_progress(progress,progress_total,1);
				swiDelay(0x10000);
			}
		}
	}
	
	text(1,"Done! You can turn off your DS.");
	text(2,"                               ");
	while(1);
}
